@@ -25,7 +25,7 @@ import os.path
 
 from udiskie.common import Emitter, samefile
 from udiskie.compat import filter
-from udiskie.dbus import DBusProxy, DBusService
+from udiskie.dbus import DBusProxy, DBusService, DBusException
 
 
 __all__ = ['Sniffer', 'Daemon']
@@ -42,6 +42,8 @@ class DeviceBase(object):
 
     Interface = 'org.freedesktop.UDisks.Device'
 
+    Exception = DBusException
+
     # string representation
     def __str__(self):
         """Display as object path."""
@@ -66,7 +68,7 @@ class DeviceBase(object):
             samefile(path, mp) for mp in self.mount_paths)
 
 
-class OnlineDevice(DBusProxy, DeviceBase):
+class OnlineDevice(DeviceBase):
 
     """
     Online wrapper for org.freedesktop.UDisks.Device DBus API proxy objects.
@@ -79,13 +81,14 @@ class OnlineDevice(DBusProxy, DeviceBase):
     """
 
     # construction
-    def __init__(self, udisks, proxy):
+    def __init__(self, udisks, object):
         """
         Initialize an instance with the given DBus proxy object.
 
-        proxy must be an object acquired by a call to bus.get_object().
+        :param DBusObject object:
         """
-        super(OnlineDevice, self).__init__(proxy, self.Interface)
+        self._proxy = object.get_interface(self.Interface)
+        self.object_path = object.object_path
         self.udisks = udisks
 
     # availability of interfaces
@@ -93,7 +96,7 @@ class OnlineDevice(DBusProxy, DeviceBase):
     def is_valid(self):
         """Check if there is a valid DBus object for this object path."""
         try:
-            self.property.DeviceFile
+            self._proxy.property.DeviceFile
             return True
         except self.Exception:
             return False
@@ -101,7 +104,7 @@ class OnlineDevice(DBusProxy, DeviceBase):
     @property
     def is_drive(self):
         """Check if the device is a drive."""
-        return self.property.DeviceIsDrive
+        return self._proxy.property.DeviceIsDrive
 
     @property
     def is_block(self):
@@ -111,12 +114,12 @@ class OnlineDevice(DBusProxy, DeviceBase):
     @property
     def is_partition_table(self):
         """Check if the device is a partition table."""
-        return self.property.DeviceIsPartitionTable
+        return self._proxy.property.DeviceIsPartitionTable
 
     @property
     def is_partition(self):
         """Check if the device has a partition slave."""
-        return self.property.DeviceIsPartition
+        return self._proxy.property.DeviceIsPartition
 
     @property
     def is_filesystem(self):
@@ -126,7 +129,7 @@ class OnlineDevice(DBusProxy, DeviceBase):
     @property
     def is_luks(self):
         """Check if the device is a LUKS container."""
-        return self.property.DeviceIsLuks
+        return self._proxy.property.DeviceIsLuks
 
     #----------------------------------------
     # Drive
@@ -138,26 +141,28 @@ class OnlineDevice(DBusProxy, DeviceBase):
     @property
     def is_detachable(self):
         """Check if the drive that owns this device can be detached."""
-        return self.property.DriveCanDetach if self.is_drive else None
+        return self._proxy.property.DriveCanDetach if self.is_drive else None
 
     @property
     def is_ejectable(self):
         """Check if the drive that owns this device can be ejected."""
-        return self.property.DriveIsMediaEjectable if self.is_drive else None
+        return self._proxy.property.DriveIsMediaEjectable if self.is_drive else None
 
     @property
     def has_media(self):
         """Check if there is media available in the drive."""
-        return self.property.DeviceIsMediaAvailable
+        return self._proxy.property.DeviceIsMediaAvailable
 
     # Drive methods
     def eject(self, unmount=False):
         """Eject media from the device."""
-        return self.method.DriveEject(filter_opt({'unmount': unmount}))
+        return self._proxy.method.DriveEject(
+            '(as)',
+            filter_opt({'unmount': unmount}))
 
     def detach(self):
         """Detach the device by e.g. powering down the physical port."""
-        return self.method.DriveDetach([])
+        return self._proxy.method.DriveDetach('(as)', [])
 
     #----------------------------------------
     # Block
@@ -167,19 +172,19 @@ class OnlineDevice(DBusProxy, DeviceBase):
     @property
     def device_file(self):
         """The filesystem path of the device block file."""
-        return os.path.normpath(self.property.DeviceFile)
+        return os.path.normpath(self._proxy.property.DeviceFile)
 
     @property
     def device_presentation(self):
         """The device file path to present to the user."""
-        return self.property.DeviceFilePresentation
+        return self._proxy.property.DeviceFilePresentation
 
     # TODO: device_size missing
 
     @property
     def id_usage(self):
         """Device usage class, for example 'filesystem' or 'crypto'."""
-        return self.property.IdUsage
+        return self._proxy.property.IdUsage
 
     @property
     def is_crypto(self):
@@ -189,7 +194,7 @@ class OnlineDevice(DBusProxy, DeviceBase):
     @property
     def is_ignored(self):
         """Check if the device should be ignored."""
-        return self.property.DevicePresentationHide
+        return self._proxy.property.DevicePresentationHide
 
     @property
     def id_type(self):
@@ -201,27 +206,27 @@ class OnlineDevice(DBusProxy, DeviceBase):
         IdUsage     'filesystem'    'crypto'
         IdType      'ext4'          'crypto_LUKS'
         """
-        return self.property.IdType
+        return self._proxy.property.IdType
 
     @property
     def id_label(self):
         """Label of the device if available."""
-        return self.property.IdLabel
+        return self._proxy.property.IdLabel
 
     @property
     def id_uuid(self):
         """Device UUID."""
-        return self.property.IdUuid
+        return self._proxy.property.IdUuid
 
     @property
     def luks_cleartext_slave(self):
         """Get luks crypto device."""
-        return self.udisks[self.property.LuksCleartextSlave] if self.is_luks_cleartext else None
+        return self.udisks[self._proxy.property.LuksCleartextSlave] if self.is_luks_cleartext else None
 
     @property
     def is_luks_cleartext(self):
         """Check whether this is a luks cleartext device."""
-        return self.property.DeviceIsLuksCleartext
+        return self._proxy.property.DeviceIsLuksCleartext
 
     @property
     def is_external(self):
@@ -231,7 +236,7 @@ class OnlineDevice(DBusProxy, DeviceBase):
     @property
     def is_systeminternal(self):
         """Check if the device is internal."""
-        return self.property.DeviceIsSystemInternal
+        return self._proxy.property.DeviceIsSystemInternal
 
     @property
     def drive(self):
@@ -257,7 +262,7 @@ class OnlineDevice(DBusProxy, DeviceBase):
     @property
     def partition_slave(self):
         """Get the partition slave (container)."""
-        return self.udisks[self.property.PartitionSlave] if self.is_partition else None
+        return self.udisks[self._proxy.property.PartitionSlave] if self.is_partition else None
 
     #----------------------------------------
     # Filesystem
@@ -267,14 +272,14 @@ class OnlineDevice(DBusProxy, DeviceBase):
     @property
     def is_mounted(self):
         """Check if the device is mounted."""
-        return self.property.DeviceIsMounted
+        return self._proxy.property.DeviceIsMounted
 
     @property
     def mount_paths(self):
         """Return list of active mount paths."""
         if not self.is_mounted:
             return []
-        raw_paths = self.property.DeviceMountPaths
+        raw_paths = self._proxy.property.DeviceMountPaths
         return [os.path.normpath(path) for path in raw_paths]
 
     # Filesystem methods
@@ -286,11 +291,16 @@ class OnlineDevice(DBusProxy, DeviceBase):
         options = (options or []) + filter_opt({
             'auth_no_user_interaction': auth_no_user_interaction
         })
-        return self.method.FilesystemMount(fstype or self.id_type, options)
+        return self._proxy.method.FilesystemMount(
+            '(sas)',
+            fstype or self.id_type,
+            options)
 
     def unmount(self, force=False):
         """Unmount filesystem."""
-        return self.method.FilesystemUnmount(filter_opt({'force': force}))
+        return self._proxy.method.FilesystemUnmount(
+            '(as)',
+            filter_opt({'force': force}))
 
     #----------------------------------------
     # Encrypted
@@ -300,7 +310,7 @@ class OnlineDevice(DBusProxy, DeviceBase):
     @property
     def luks_cleartext_holder(self):
         """Get unlocked luks cleartext device."""
-        return self.udisks[self.property.LuksHolder] if self.is_luks else None
+        return self.udisks[self._proxy.property.LuksHolder] if self.is_luks else None
 
     @property
     def is_unlocked(self):
@@ -310,11 +320,15 @@ class OnlineDevice(DBusProxy, DeviceBase):
     # Encrypted methods
     def unlock(self, password):
         """Unlock Luks device."""
-        return self.udisks.update(self.method.LuksUnlock(password, []))
+        return self.udisks.update(
+            self._proxy.method.LuksUnlock(
+                '(sas)',
+                password,
+                []))
 
     def lock(self):
         """Lock Luks device."""
-        return self.method.LuksLock([])
+        return self._proxy.method.LuksLock('(as)', [])
 
     #----------------------------------------
     # derived properties
@@ -366,6 +380,8 @@ class CachedDevice(DeviceBase):
 
     def __getattr__(self, key):
         """Resolve unknown properties and methods via the online device."""
+        if key.startswith('_'):
+            raise AttributeError(key)
         return getattr(self._device, key)
 
     # Overload properties that return Device objects to return CachedDevice
@@ -456,20 +472,12 @@ class Sniffer(UDisks):
 
     def get(self, object_path):
         """Create a Device instance from object path."""
-        return OnlineDevice(self, self._proxy._bus.get_object(self.BusName,
-                                                              object_path))
+        return OnlineDevice(
+            self,
+            self._proxy.object.bus.get_object(object_path))
     update = get
 
 
-class Job(object):
-
-    """Job information struct for devices."""
-
-    def __init__(self, job_id, percentage):
-        self.job_id = job_id
-        self.percentage = percentage
-
-
 class Daemon(Emitter, UDisks):
 
     """
@@ -489,33 +497,28 @@ class Daemon(Emitter, UDisks):
     `disconnect` can be used to add or remove event handlers.
     """
 
-    mainloop = True
-
     def __init__(self, proxy=None):
         """
         Create a Daemon object and start listening to DBus events.
 
         :param common.DBusProxy proxy: proxy to the dbus service object
-        :param udisks1.Sniffer sniffer: sniffer to use
 
-        If neither proxy nor sniffer are given they will be created and
-        dbus will be configured for the gobject mainloop.
+        A default proxy will be created if set to ``None``.
         """
-        event_names = [stem + suffix
-                       for suffix in ('ed', 'ing')
-                       for stem in (
-                           'device_add',
-                           'device_remov',
-                           'device_mount',
-                           'device_unmount',
-                           'media_add',
-                           'media_remov',
-                           'device_unlock',
-                           'device_lock',
-                           'device_chang', )] + ['job_failed']
+        event_names = ['device_added',
+                       'device_removed',
+                       'device_mounted',
+                       'device_unmounted',
+                       'media_added',
+                       'media_removed',
+                       'device_unlocked',
+                       'device_locked',
+                       'device_changed',
+                       'job_failed']
         super(Daemon, self).__init__(event_names)
 
-        sniffer = Sniffer(proxy or self.connect_service())
+        proxy = proxy or self.connect_service()
+        sniffer = Sniffer(proxy)
 
         self._sniffer = sniffer
         self._jobs = {}
@@ -525,23 +528,10 @@ class Daemon(Emitter, UDisks):
                         'eject': {}, 'detach': {}}
 
         self.connect('device_changed', self._on_device_changed)
-        bus = self._sniffer._proxy._bus
-        bus.add_signal_receiver(
-            self._device_added,
-            signal_name='DeviceAdded',
-            bus_name=self.BusName)
-        bus.add_signal_receiver(
-            self._device_removed,
-            signal_name='DeviceRemoved',
-            bus_name=self.BusName)
-        bus.add_signal_receiver(
-            self._device_changed,
-            signal_name='DeviceChanged',
-            bus_name=self.BusName)
-        bus.add_signal_receiver(
-            self._device_job_changed,
-            signal_name='DeviceJobChanged',
-            bus_name=self.BusName)
+        proxy.connect('DeviceAdded', self._device_added)
+        proxy.connect('DeviceRemoved', self._device_removed)
+        proxy.connect('DeviceChanged', self._device_changed)
+        proxy.connect('DeviceJobChanged', self._device_job_changed)
         self._sync()
 
     # Sniffer overrides
@@ -571,12 +561,30 @@ class Daemon(Emitter, UDisks):
     # events
     def _on_device_changed(self, old_state, new_state):
         """Detect type of event and trigger appropriate event handlers."""
-        d = {}
-        d['media_added'] = new_state.has_media and not old_state.has_media
-        d['media_removed'] = old_state.has_media and not new_state.has_media
-        for event in d:
-            if d[event]:
-                self.trigger(event, new_state)
+        self._detect_toggle('has_media', old_state, new_state,
+                            'media_added', 'media_removed')
+        self._detect_toggle('is_mounted', old_state, new_state,
+                            'device_mounted', 'device_unmounted')
+        self._detect_toggle('is_unlocked', old_state, new_state,
+                            'device_unlocked', 'device_locked')
+
+    def _detect_toggle(self, property_name, old, new, add_name, del_name):
+        old_valid = old and bool(getattr(old, property_name))
+        new_valid = new and bool(getattr(new, property_name))
+        # If we were notified about a started job we don't want to trigger
+        # an event when the device is changed, but when the job is
+        # completed. Otherwise we would show unmount notifications too
+        # early (when it's not yet safe to remove the drive).
+        # On the other hand, if the unmount operation is not issued via
+        # UDisks1, there will be no corresponding job.
+        cached_job = self._jobs.get(old.object_path)
+        action_name = self._event_mapping.get(cached_job)
+        if add_name and new_valid and not old_valid:
+            if add_name != action_name:
+                self.trigger(add_name, new)
+        elif del_name and old_valid and not new_valid:
+            if del_name != action_name:
+                self.trigger(del_name, new)
 
     # UDisks event listeners
     def _device_added(self, object_path):
@@ -610,27 +618,32 @@ class Daemon(Emitter, UDisks):
 
         Internal method.
         """
-        if not job_in_progress and object_path in self._jobs:
-            job_id = self._jobs[object_path].job_id
         try:
-            action = self._action_mapping[job_id]
+            if job_id:
+                action = self._action_mapping[job_id]
+            else:
+                action = self._jobs[object_path]
         except KeyError:
+            # this can happen
+            # a) at startup, when we only see the completion of a job
+            # b) when we get notified about a job, which we don't handle
             return
-        event_name = self._event_mapping[action]
-        dev = self[object_path]
         # NOTE: The here used heuristic is prone to raise conditions.
         if job_in_progress:
-            self.trigger(event_name + 'ing', dev, job_percentage)
-            self._jobs[object_path] = Job(job_id, job_percentage)
-        elif self._check_success[job_id](dev):
-            self.trigger(event_name + 'ed', dev)
-            del self._jobs[object_path]
+            # Cache the action name for later use:
+            self._jobs[object_path] = action
         else:
-            # get and delete message, if available:
-            message = self._errors[action].pop(object_path, "")
-            self.trigger('job_failed', dev, 'device_' + action, message)
-            log = logging.getLogger(__name__)
-            log.info('%s operation failed for device: %s' % (job_id, object_path))
+            del self._jobs[object_path]
+            device = self[object_path]
+            if self._check_success[action](device):
+                event = self._event_mapping[action]
+                self.trigger(event, device)
+            else:
+                # get and delete message, if available:
+                message = self._errors[action].pop(object_path, "")
+                self.trigger('job_failed', device, action, message)
+                log = logging.getLogger(__name__)
+                log.info('%s operation failed for device: %s' % (job_id, object_path))
 
     # used internally by _device_job_changed:
     _action_mapping = {
@@ -639,31 +652,32 @@ class Daemon(Emitter, UDisks):
         'LuksUnlock': 'unlock',
         'LuksLock': 'lock',
         'DriveDetach': 'detach',
-        'DriveEject': 'eject' }
+        'DriveEject': 'eject'}
 
-    _event_mapping = {'mount': 'device_mount',
-                      'unmount': 'device_unmount',
-                      'unlock': 'device_unlock',
-                      'lock': 'device_lock',
-                      'eject': 'media_remov',
-                      'detach': 'device_remov'}
+    _event_mapping = {
+        'mount': 'device_mounted',
+        'unmount': 'device_unmounted',
+        'unlock': 'device_unlocked',
+        'lock': 'device_locked',
+        'eject': 'media_removed',
+        'detach': 'device_removed'}
 
     _check_success = {
-        'FilesystemMount': lambda dev: dev.is_mounted,
-        'FilesystemUnmount': lambda dev: not dev or not dev.is_mounted,
-        'LuksUnlock': lambda dev: dev.is_unlocked,
-        'LuksLock': lambda dev: not dev or not dev.is_unlocked,
-        'DriveDetach': lambda dev: not dev,
-        'DriveEject': lambda dev: not dev or not dev.has_media
+        'mount': lambda dev: dev.is_mounted,
+        'unmount': lambda dev: not dev or not dev.is_mounted,
+        'unlock': lambda dev: dev.is_unlocked,
+        'lock': lambda dev: not dev or not dev.is_unlocked,
+        'detach': lambda dev: not dev,
+        'eject': lambda dev: not dev or not dev.has_media
     }
 
     # internal state keeping
     def _sync(self):
         """Cache all device states."""
-        self._devices = { dev.object_path: dev for dev in self._sniffer }
+        online_devices = { dev.object_path: dev for dev in self._sniffer }
         self._devices = {
             object_path: CachedDevice(device)
-            for object_path,device in self._devices.items() }
+            for object_path,device in online_devices.items() }
 
     def _invalidate(self, object_path):
         """Flag the device invalid. This removes it from the iteration."""
